// Copyright (c) 2002 Andreas Jnsson, www.AngelCode.com

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "winhook.h"

// Global variables
HINSTANCE hDllModule = 0;

// We will share this data between all instances of the DLL
#pragma data_seg(".HOOKDATA")
HWND hookWnd              = 0;
HWND ownerWnd             = 0;
HHOOK hHookCallWndProc    = 0;
HHOOK hHookGetMessage     = 0;
HHOOK hHookCallWndProcRet = 0;
#pragma data_seg()
#pragma comment(linker, "/SECTION:.HOOKDATA,RWS")


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			// Store the DLL's instance handle
			hDllModule = (HINSTANCE)hModule;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

// Sets up the hooks needed
WINHOOK_API int SetHook(HWND _hookWnd, HWND _ownerWnd)
{
	if( hookWnd || ownerWnd )
		return -1;

	hookWnd = _hookWnd;
	ownerWnd = _ownerWnd;

	// We want to catch posted WM_PAINT messages before they are processed
	hHookGetMessage = SetWindowsHookEx(WH_GETMESSAGE, (HOOKPROC)GetMessageHook, hDllModule, 0); 
	if( hHookGetMessage == 0 ) 
		return -1;

	// and sent WM_PAINT messages before they are processed
	hHookCallWndProc = SetWindowsHookEx(WH_CALLWNDPROC, (HOOKPROC)CallWndProcHook, hDllModule, 0); 
	if( hHookCallWndProc == 0 ) 
		return -1;

	// and sent WM_ERASEBKGND messages after they have been processed
	hHookCallWndProcRet = SetWindowsHookEx(WH_CALLWNDPROCRET, (HOOKPROC)CallWndProcRetHook, hDllModule, 0); 
	if( hHookCallWndProcRet == 0 ) 
		return -1;

	return 0;
}

// Remove the hooks
WINHOOK_API int RemoveHook()
{
	if( hHookGetMessage )
		UnhookWindowsHookEx(hHookGetMessage);

	if( hHookCallWndProc )
		UnhookWindowsHookEx(hHookCallWndProc);

	if( hHookCallWndProcRet )
		UnhookWindowsHookEx(hHookCallWndProcRet);

	hookWnd = 0;
	ownerWnd = 0;

	return 0;
}

LRESULT CALLBACK GetMessageHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// if nCode is negative we are not allowed to do anything
	if( nCode >= 0 && wParam == PM_REMOVE )
	{
		// We only want to process WM_PAINT if it was removed from the queue
		HWND hWnd = ((MSG*)lParam)->hwnd;
		UINT message = ((MSG*)lParam)->message;
		if( hWnd == hookWnd && message == WM_PAINT )
		{
			// Just in case the main program crashes we'll 
			// make an automatic removal of the hooks
			if( IsWindow(ownerWnd) )
			{
				// We'll send the message with a timeout because
				// we want to make sure the main program has time
				// to finish drawing before we let the desktop 
				// window continue
				DWORD result;
				SendMessageTimeout(ownerWnd, WM_APP+100, (WPARAM)WM_PAINT, 1, SMTO_BLOCK, 500, &result);
			}
			else
				RemoveHook();
		}
	}

	return CallNextHookEx(hHookGetMessage, nCode, wParam, lParam);
}


LRESULT CALLBACK CallWndProcHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// if nCode is negative we are not allowed to do anything
	if( nCode >= 0 )
	{
		// We want to process WM_PAINT before it enters the window procedure
		HWND hWnd = ((CWPSTRUCT*)lParam)->hwnd;
		UINT message = ((CWPSTRUCT*)lParam)->message;
		if( hWnd == hookWnd && message == WM_PAINT )
		{
			// Just in case the main program crashes we'll 
			// make an automatic removal of the hooks
			if( IsWindow(ownerWnd) )
			{
				// We'll send the message with a timeout because
				// we want to make sure the main program has time
				// to finish drawing before we let the desktop 
				// window continue
				DWORD result;
				SendMessageTimeout(ownerWnd, WM_APP+100, (WPARAM)WM_PAINT, 1, SMTO_BLOCK, 500, &result);
			}
			else
				RemoveHook();
		}
	}

	return CallNextHookEx(hHookCallWndProc, nCode, wParam, lParam);
}

LRESULT CALLBACK CallWndProcRetHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// if nCode is negative we are not allowed to do anything
	if( nCode >= 0 )
	{
		// We only want WM_ERASEBKGND after it has finished
		HWND hWnd = ((CWPRETSTRUCT*)lParam)->hwnd;
		UINT message = ((CWPRETSTRUCT*)lParam)->message;
		if( hWnd == hookWnd && message == WM_ERASEBKGND )
		{
			// Just in case the main program crashes we'll 
			// make an automatic removal of the hooks
			if( IsWindow(ownerWnd) )
			{
				// We'll send the message with a timeout because
				// we want to make sure the main program has time
				// to finish drawing before we let the desktop 
				// window continue
				DWORD result;
				SendMessageTimeout(ownerWnd, WM_APP+100, (WPARAM)WM_ERASEBKGND, 0, SMTO_BLOCK, 500, &result);
			}
			else
				RemoveHook();
		}
	}

	return CallNextHookEx(hHookCallWndProcRet, nCode, wParam, lParam);
}

